/************************************************************************
 *
 * copyup_c.cpp
 *
 ************************************************************************
 *
 * screen copyup management code
 *
 ************************************************************************/

/*
 * includes
 */

#include "std.h"
#include "mem.h"
#include "copyup.h"
#include "main.h"
#include "gba.h"
#include "emu.h"
#include "font.h"
#include "int.h"
#include "shell.h"

/*
 * defines
 */

/*
 * structures/classes
 */

/*
 * globals
 */

byte		*codeBase;
uint32		codeEntry;
uint32		*tableBase;
uint32		*slowTableBase;
int32		copyupMode;
byte		screenNo;
uint32		*sm1,*sm2;
uint16		basePalette[256];
int32		scrollOffsetX,scrollOffsetY;

/*
 * prototypes
 */

/*
 * code
 */

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

void CopyupInit(void)
{
	int32		i,j;

	screenNo=0;

	// find largest copyup method size
	i=(uint32)(&AsmCopyScreenLinearEnd) - (uint32)(&AsmCopyScreenLinearStart);
	j=(uint32)(&AsmCopyScreenTiledScaledEnd) - (uint32)(&AsmCopyScreenTiledScaledStart);
	if(j>i)	i=j;
	j=(uint32)(&AsmCopyScreenTiledScaled256End) - (uint32)(&AsmCopyScreenTiledScaled256Start);
	if(j>i)	i=j;
	j=(uint32)(&AsmCopyScreenTiledEnd) - (uint32)(&AsmCopyScreenTiledStart);
	if(j>i)	i=j;

	// allocate
	codeBase=(byte*)MemAlloc(HEAP_INTERNAL,i);
	ASSERT(codeBase);

	// allocate 2K table
	tableBase=(uint32 *)MemAlloc(HEAP_INTERNAL,1024*2);
	ASSERT(tableBase);

	// allocate slow 2K table
	slowTableBase=(uint32 *)MemAlloc(HEAP_EXTERNAL,1024*2);
	ASSERT(slowTableBase);
}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

void CopyupSetMode(int32 mode)
{
	int32		i,j;

	copyupMode=mode;

	// disable Y-match interrupts and H-blank interrupts by default!
	*((uint16*)(REG_BASE+0x004))=0x0008;

	// no scaling
	*(uint16 *)(REG_BASE+0x020) = 0x100;	// horiz scale
	*(uint16 *)(REG_BASE+0x026) = 0x100;	// vert scale

	//
	// init palette and colour lookup tables
	//
	if(mode==CM_LINEAR || mode==CM_TILEDSCALED256 || mode==CM_TILEDSCALED256_INTERLACED)
	{
		// init palette
		for(i=0;i<256;i++)
		{
			int32		r,g,b,index;

			if(i&128)	index=(i>>0)&7;
			else		index=(i>>3)&7;

			r=0;	g=0;	b=0;
			if(index&1)	b+=0x80;
			if(index&2)	r+=0x80;
			if(index&4)	g+=0x80;
			if(i&64)	{ b*=2; g*=2; r*=2; }
			if(r>0xFF) r=0xFF;
			if(g>0xFF) g=0xFF;
			if(b>0xFF) b=0xFF;

			r>>=3;	g>>=3;	b>>=3;
			j=(b<<10)|(g<<5)|(r<<0);
			basePalette[i]=j;
		}


		// init exploding table
		for(i=0;i<256;i++)
		{
			uint32		d,e;

			d=0;	e=0;
			if(i&128)	d|=(1<<7);
			if(i&64)	d|=(1<<15);
			if(i&32)	d|=(1<<23);
			if(i&16)	d|=(1<<31);
			if(i&8)		e|=(1<<7);
			if(i&4)		e|=(1<<15);
			if(i&2)		e|=(1<<23);
			if(i&1)		e|=(1<<31);
			tableBase[i*2+0]=d;
			tableBase[i*2+1]=e;
		}
	
	}
	else if(mode==CM_TILEDSCALED || mode==CM_TILEDSCALED_INTERLACED || mode==CM_TILED || mode==CM_TILED_LOCKED)
	{
		// init palette
		for(i=0;i<16;i++)
		{
			int32		r,g,b;

			r=0;	g=0;	b=0;
			if(i&1)	b+=0x80;
			if(i&2)	r+=0x80;
			if(i&4)	g+=0x80;
			if(i&8)	{ b*=2; g*=2; r*=2; }
			if(r>0xFF) r=0xFF;
			if(g>0xFF) g=0xFF;
			if(b>0xFF) b=0xFF;

			r>>=3;	g>>=3;	b>>=3;
			j=(b<<10)|(g<<5)|(r<<0);
			basePalette[i]=j;
		}

		// init exploding table
		for(i=0;i<256;i++)
		{
			uint32		d;

			d=0;
			for(j=0;j<8;j++)
			{
				if(i & (1<<(7-j)))	d |= 0xF << (j*4);
			}
			tableBase[i]=d;
		}

		// init ATTR -> INK/PAPER tables
		for(i=0;i<256;i++)
		{
			uint32		d,e;

			d=(i&7);	if(i&64)	d|=8;
			e=(i>>3)&7;	if(i&64)	e|=8;
			d|=d<<4;	d|=d<<8;	d|=d<<16;
			e|=e<<4;	e|=e<<8;	e|=e<<16;
			slowTableBase[i]=d;
			tableBase[i+256]=e;
		}
	}

	//
	// setup video hardware, copy up code
	//
	if(mode==CM_LINEAR)
	{
		// set linear mode
		*(uint16 *)REG_LCD = LCD_MODE4 | LCD_BG2_EN;
	}
	else if(mode==CM_TILEDSCALED || mode==CM_TILEDSCALED_INTERLACED)
	{
		int32		k,m;

		// init tiled mode
		*(uint16 *)REG_LCD = LCD_MODE0 | LCD_BG0_EN;
		*(uint16 *)REG_BG0 = 0x0004;		// 16-colour, chardata at 16384, tilemap at 0
		*(uint16 *)(REG_BASE+0x010) = 0x0008;	// horiz scrolling!
		*(uint16 *)(REG_BASE+0x012) = 0x0004;	// vert scrolling

		// init tilemap
		m=0;
		for(i=0;i<3;i++)
		{
			for(k=0;k<7;k++)
			{
				for(j=0;j<32;j++)
				{
					((uint16*)VRAM)[m]=(i*7*32)+(j*7)+k;
					m++;
				}
			}
		}

	}
	else if(mode==CM_TILED || mode==CM_TILED_LOCKED)
	{
		int32		k,m;

		// init tiled mode
		*(uint16 *)REG_LCD = LCD_MODE0 | LCD_BG0_EN;
		*(uint16 *)REG_BG0 = 0x0004;		// 16-colour, chardata at 16384, tilemap at 0

		// init tilemap
		m=0;
		for(i=0;i<3;i++)
		{
			for(k=0;k<8;k++)
			{
				for(j=0;j<32;j++)
				{
					((uint16*)VRAM)[m]=(i*8*32)+(j*8)+k;
					m++;
				}
			}
		}

	}
	else if(mode==CM_TILEDSCALED256 || mode==CM_TILEDSCALED256_INTERLACED)
	{
		int32		k,m;

		*(uint16 *)REG_LCD = LCD_MODE1 | LCD_BG2_EN;
		*(uint16 *)REG_BG2 = 0x4084;		// 256-colour, chardata at 16384, tilemap at 0
		*(uint16 *)(REG_BASE+0x020) = 0x100*32/30;	// horiz scale
		*(uint16 *)(REG_BASE+0x026) = 0x100*24/20;	// vert scale

		*(uint16 *)REG_BG2 = 0x408C;		// 256-colour, chardata at 32768, tilemap at 0

		// init tilemap
		m=0;
		for(i=0;i<3;i++)
		{
			for(k=0;k<8;k++)
			{
				for(j=0;j<32;j+=2)
				{
					((uint16*)VRAM)[m]=((j*8+k)<<0) | (((j+1)*8+k)<<8);
					m++;
				}
			}
		}

		// set up Y line match values
		intArgs->yMatchId=0;
		intArgs->yMatchLine0=0xD228;		intArgs->yMatchReg0=0x408C;
		intArgs->yMatchLine1=0x3528;		intArgs->yMatchReg1=0x4084;
		intArgs->yMatchLine2=0x6A28;		intArgs->yMatchReg2=0x4088;

		// enable Y match interrupts
		*((uint16*)(REG_BASE+0x004))=0x0028;

	}

	//
	// copy up code
	//
	if(copyupMode==CM_LINEAR)
	{
		// copy code
		mymemcpy(codeBase,&AsmCopyScreenLinearStart,(uint32)(&AsmCopyScreenLinearEnd) - (uint32)(&AsmCopyScreenLinearStart));
		codeEntry=(uint32)codeBase;
		codeEntry+=(uint32)(&AsmCopyScreenLinearEntry) - (uint32)(&AsmCopyScreenLinearStart);
	}
	else if(copyupMode==CM_TILED || copyupMode==CM_TILED_LOCKED)
	{
		// copy code
		mymemcpy(codeBase,&AsmCopyScreenTiledStart,(uint32)(&AsmCopyScreenTiledEnd) - (uint32)(&AsmCopyScreenTiledStart));
		codeEntry=(uint32)codeBase;
		codeEntry+=(uint32)(&AsmCopyScreenTiledEntry) - (uint32)(&AsmCopyScreenTiledStart);
	}
	else if(copyupMode==CM_TILEDSCALED || copyupMode==CM_TILEDSCALED_INTERLACED)
	{
		// copy code
		mymemcpy(codeBase,&AsmCopyScreenTiledScaledStart,(uint32)(&AsmCopyScreenTiledScaledEnd) - (uint32)(&AsmCopyScreenTiledScaledStart));
		codeEntry=(uint32)codeBase;
		codeEntry+=(uint32)(&AsmCopyScreenTiledScaledEntry) - (uint32)(&AsmCopyScreenTiledScaledStart);

		// get self-modifying code ptrs
		sm1=(uint32*)(codeBase + ((uint32)(&acstssm1) - (uint32)(&AsmCopyScreenTiledScaledStart)));
		sm2=(uint32*)(codeBase + ((uint32)(&acstssm2) - (uint32)(&AsmCopyScreenTiledScaledStart)));
	}
	else if(copyupMode==CM_TILEDSCALED256 || copyupMode==CM_TILEDSCALED256_INTERLACED)
	{
		// copy code
		mymemcpy(codeBase,&AsmCopyScreenTiledScaled256Start,(uint32)(&AsmCopyScreenTiledScaled256End) - (uint32)(&AsmCopyScreenTiledScaled256Start));
		codeEntry=(uint32)codeBase;
		codeEntry+=(uint32)(&AsmCopyScreenTiledScaled256Entry) - (uint32)(&AsmCopyScreenTiledScaled256Start);
	}

	// mess with brightness/bw
	for(i=0;i<256;i++)
	{
		int32		r,g,b;
		uint16		v;

		v=basePalette[i];
		r=(v>>0)&31;	g=(v>>5)&31;	b=(v>>10)&31;
		r<<=3; g<<=3; b<<=3;
		if(colourMode==COL_BW)	{ r=(r+g+b)/3; g=r; b=r; }
		if(brightness==1)	{ r=r*3/2; g=g*3/2; b=b*3/2; }
		if(brightness==2)	{ r=r*7/4; g=g*7/4; b=b*7/4; }
		if(r>255) r=255;
		if(g>255) g=255;
		if(b>255) b=255;
		r>>=3; g>>=3; b>>=3;
		v=(r<<0) | (g<<5) | (b<<10);
		((uint16*)PALETTE)[i]=v;
	}
}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

void DoCopyup(void)
{
	int32			i,k,l;
	CAsmScreenCopyArgs	args;

	// set parm ptr
	((uint32*)(codeBase))[0]=(uint32)(&args);

	if(copyupMode==CM_LINEAR)
	{
		// copy 3 pels of top line
		args.gbaDest=(uint32)VRAM;
		args.specAttSrc=1+(uint32)(&z80Mem[0x5800]);
		args.specBmpSrc=1+(uint32)(&z80Mem[0x4000])+(5*256);
		args.lineCount=3;
		args.bmpTable=(uint32)tableBase;
		args.colCount=30;
		Call32(codeEntry);

		for(i=1;i<24;i++)
		{
			k=i>>3;
			l=i&7;

			args.gbaDest=(uint32)VRAM + ((i*7-4)*240);
			args.specAttSrc=1+(uint32)(&z80Mem[0x5800+i*32]);
			args.specBmpSrc=1+(uint32)(&z80Mem[0x4000+(l*32)+(k*2048)]);
			if(i==23)	args.lineCount=3;
			else		args.lineCount=7;
			Call32(codeEntry);
		}
	}
	else if(copyupMode==CM_TILEDSCALED || copyupMode==CM_TILEDSCALED_INTERLACED ||
		copyupMode==CM_TILEDSCALED256 || copyupMode==CM_TILEDSCALED256_INTERLACED ||
		copyupMode==CM_TILED || copyupMode==CM_TILED_LOCKED)
	{
		// set self-modifying code!
		if(copyupMode==CM_TILEDSCALED)
		{
			*sm1=0xe4d1a100;
			*sm2=0xe4d1a200;
		}
		else if(copyupMode==CM_TILEDSCALED_INTERLACED)
		{
			screenNo^=1;
			if(copyupMode!=CM_TILEDSCALED_INTERLACED)	screenNo=0;

			if(screenNo&1)
			{
				*sm1=0xe4d1a200;
				*sm2=0xe4d1a100;
			}
			else
			{
				*sm1=0xe4d1a100;
				*sm2=0xe4d1a200;
			}
		}

		args.bmpTable=(uint32)tableBase;
		args.attTable=(uint32)slowTableBase;
		args.colCount=32;
		args.lineCount=8;
		
		if(copyupMode==CM_TILEDSCALED || copyupMode==CM_TILEDSCALED_INTERLACED)
		{
			for(i=0;i<3;i++)
			{
				args.gbaDest=VRAM + 16384 + (i*7*32*32);
				args.specAttSrc=(uint32)(&z80Mem[0x5800 + (i*256)]);
				args.specBmpSrc=(uint32)(&z80Mem[0x4000 + (i*2048)]);
				Call32(codeEntry);
			}
		}
		else if(copyupMode==CM_TILED || copyupMode==CM_TILED_LOCKED)
		{
			for(i=0;i<3;i++)
			{
				args.gbaDest=VRAM + 16384 + (i*8*32*32);
				args.specAttSrc=(uint32)(&z80Mem[0x5800 + (i*256)]);
				args.specBmpSrc=(uint32)(&z80Mem[0x4000 + (i*2048)]);
				Call32(codeEntry);
			}
		}
		else
		{
			for(i=0;i<3;i++)
			{
				args.gbaDest=VRAM + 16384 + (i*8*32*64);
				args.specAttSrc=(uint32)(&z80Mem[0x5800 + (i*256)]);
				args.specBmpSrc=(uint32)(&z80Mem[0x4000 + (i*2048)]);
				Call32(codeEntry);
			}
		}
	}

	// scroll?
	if(copyupMode==CM_TILED || copyupMode==CM_TILED_LOCKED)
	{
		int32		key;

		if(copyupMode==CM_TILED)
		{
			key=*((volatile uint16*)REG_P1) ^ 0xFFFF;
			if(key & 256)
			{
				if(key&128)	scrollOffsetY++;
				if(key&64)	scrollOffsetY--;
				if(key&32)	scrollOffsetX--;
				if(key&16)	scrollOffsetX++;
			}
			if(scrollOffsetY<0)	scrollOffsetY=0;
			if(scrollOffsetY>=32)	scrollOffsetY=31;
			if(scrollOffsetX<0)	scrollOffsetX=0;
			if(scrollOffsetX>=16)	scrollOffsetX=15;
		}

		*(uint16 *)(REG_BASE+0x010) = scrollOffsetX;
		*(uint16 *)(REG_BASE+0x012) = scrollOffsetY;
	}
}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

void DoThirdCopyup(int32 which)
{
	int32			i,k,l;
	CAsmScreenCopyArgs	args;

	// set parm ptr
	((uint32*)(codeBase))[0]=(uint32)(&args);

	if(copyupMode==CM_LINEAR)
	{
		args.bmpTable=(uint32)tableBase;
		args.colCount=30;

		for(i=which*8;i<which*8+8;i++)
		{
			k=i>>3;
			l=i&7;

			if(!i)
			{
				// copy 3 pels of top line
				args.gbaDest=(uint32)VRAM;
				args.specAttSrc=1+(uint32)(&z80Mem[0x5800]);
				args.specBmpSrc=1+(uint32)(&z80Mem[0x4000])+(5*256);
				args.lineCount=3;
			}
			else if(i==23)
			{
				args.gbaDest=(uint32)VRAM + ((i*7-4)*240);
				args.specAttSrc=1+(uint32)(&z80Mem[0x5800+i*32]);
				args.specBmpSrc=1+(uint32)(&z80Mem[0x4000+(l*32)+(k*2048)]);
				args.lineCount=3;
			}
			else
			{
				args.gbaDest=(uint32)VRAM + ((i*7-4)*240);
				args.specAttSrc=1+(uint32)(&z80Mem[0x5800+i*32]);
				args.specBmpSrc=1+(uint32)(&z80Mem[0x4000+(l*32)+(k*2048)]);
				args.lineCount=8;
			}
			Call32(codeEntry);
		}
	}
	else if(copyupMode==CM_TILEDSCALED || copyupMode==CM_TILEDSCALED_INTERLACED ||
		copyupMode==CM_TILEDSCALED256 || copyupMode==CM_TILEDSCALED256_INTERLACED ||
		copyupMode==CM_TILED || copyupMode==CM_TILED_LOCKED)
	{
		// set self-modifying code!
		if(copyupMode==CM_TILEDSCALED)
		{
			*sm1=0xe4d1a100;
			*sm2=0xe4d1a200;
		}
		else if(copyupMode==CM_TILEDSCALED_INTERLACED)
		{
			if(!which)
			{
				screenNo^=1;
				if(copyupMode!=CM_TILEDSCALED_INTERLACED)	screenNo=0;
			}

			if(screenNo&1)
			{
				*sm1=0xe4d1a200;
				*sm2=0xe4d1a100;
			}
			else
			{
				*sm1=0xe4d1a100;
				*sm2=0xe4d1a200;
			}
		}

		args.bmpTable=(uint32)tableBase;
		args.attTable=(uint32)slowTableBase;
		args.colCount=32;
		args.lineCount=8;
		
		if(copyupMode==CM_TILEDSCALED || copyupMode==CM_TILEDSCALED_INTERLACED)
		{
			i=which;
			{
				args.gbaDest=VRAM + 16384 + (i*7*32*32);
				args.specAttSrc=(uint32)(&z80Mem[0x5800 + (i*256)]);
				args.specBmpSrc=(uint32)(&z80Mem[0x4000 + (i*2048)]);
				Call32(codeEntry);
			}
		}
		else if(copyupMode==CM_TILED || copyupMode==CM_TILED_LOCKED)
		{
			i=which;
			{
				args.gbaDest=VRAM + 16384 + (i*8*32*32);
				args.specAttSrc=(uint32)(&z80Mem[0x5800 + (i*256)]);
				args.specBmpSrc=(uint32)(&z80Mem[0x4000 + (i*2048)]);
				Call32(codeEntry);
			}
		}
		else
		{
			i=which;
			{
				args.gbaDest=VRAM + 16384 + (i*8*32*64);
				args.specAttSrc=(uint32)(&z80Mem[0x5800 + (i*256)]);
				args.specBmpSrc=(uint32)(&z80Mem[0x4000 + (i*2048)]);
				Call32(codeEntry);
			}
		}
	}

	// scroll?
	if(copyupMode==CM_TILED || copyupMode==CM_TILED_LOCKED)
	{
		int32		key;

		if(copyupMode==CM_TILED)
		{
			key=*((volatile uint16*)REG_P1) ^ 0xFFFF;
			if(key & 512)
			{
				if(key&128)	scrollOffsetY++;
				if(key&64)	scrollOffsetY--;
				if(key&32)	scrollOffsetX--;
				if(key&16)	scrollOffsetX++;
			}
			if(scrollOffsetY<0)	scrollOffsetY=0;
			if(scrollOffsetY>=32)	scrollOffsetY=31;
			if(scrollOffsetX<0)	scrollOffsetX=0;
			if(scrollOffsetX>=16)	scrollOffsetX=15;
		}

		*(uint16 *)(REG_BASE+0x010) = scrollOffsetX;
		*(uint16 *)(REG_BASE+0x012) = scrollOffsetY;
	}
}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

