/************************************************************************
 *
 * emu.cpp
 *
 ************************************************************************
 *
 * main emulation loop code
 *
 ************************************************************************/

//#define SHADOW
#define USE_AUDIO

/*
 * includes
 */

#include "std.h"
#include "emu.h"
#include "z80.h"
#include "font.h"
#include "mem.h"
#include "gba.h"
#include "int.h"
#include "copyup.h"
#include "main.h"
#include "shell.h"

#ifdef SHADOW
#include "shadow.h"
#endif

/*
 * defines
 */

#define FLG_CARE		0xd7

/*
 * structures/classes
 */

/*
 * globals
 */

extern byte	specRom[];

CZ80Asm*	z80Ptr;
uint32		z80Base;
uint32		z80Entry;
uint32*		jtBase;
uint32*		jtFast;
byte*		z80Mem;

uint32		flashEntry;
uint32		bkmEntry;
byte*		ports;

CZ80State	bkpState;
byte		*bkpMem;

int32		lastPC;
int32		lastOp,lastOp2;
int32		lastFlags;

bool		singleStep;
bool		trace,memFailure;
uint16		memFailAddr;

int32		frameCt;
int32		intCt;
int32		lastKey;
int32		frameSkipCt;

#ifdef SHADOW
byte		*shadowMem;
Z80		shadowZ80;
#endif

KeyMap		keys[16];
KeyMap		bkpKeys[16];
byte		keyboard[8];

uint32		imageFlags;

/*
 * prototypes
 */

int32		CheckState(void);
void		DumpState(void);
void		EmuDoInt(void);
void		EmuCopyState(void);
void		EmuBackup(void);
void		EmuRestore(void);
void		EmuCopyLine(int32 line);

/*
 * code
 */

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

void WaitKey(void)
{
	int32		key;

	do
	{
		key=*((volatile uint16*)REG_P1) ^ 0xFFFF;
	}	while(key&64);
	do
	{
		key=*((volatile uint16*)REG_P1) ^ 0xFFFF;
	}	while(!(key&64));
}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

void EmuInit(void)
{
	extern int32	Z80CORESTART;
	extern int32	Z80CoreEntry;
	extern int32	BuildKeyMasks;
	extern int32	JT_MAIN;
	extern int32	DoCB,DoED,DoDD,DoFD,DoDDCB,DoFDCB,DoBAD;
	extern int32	DoFlash;
	extern byte	rotateFlagTable[];
	int32		i;

	// set counters
	frameCt=0;
	frameSkipCt=0;
	intCt=0;
	singleStep=false;

	// get length of copyup code
	i=(uint32)(&JT_MAIN) - (uint32)(&Z80CORESTART);

	// allocate memory in internal RAM
	z80Base=(uint32)MemAlloc(HEAP_INTERNAL,i);
	ASSERT(z80Base);

	// copy Z80CORESTART up to internal RAM (ROM image already includes space for Z80 structures)
	memcpy((byte*)z80Base,&Z80CORESTART,i);
	z80Ptr=(CZ80Asm*)z80Base;
	for(i=0;i<(int32)sizeof(CZ80Asm);i++) ((byte*)z80Ptr)[i]=0;
	z80Ptr->statePtr=&z80Ptr->state;

	// compute code entry point(s)
	z80Entry=z80Base + ((uint32)&Z80CoreEntry - (uint32)&Z80CORESTART);
	flashEntry=z80Base + ((uint32)&DoFlash - (uint32)&Z80CORESTART);
	bkmEntry=z80Base + ((uint32)&BuildKeyMasks - (uint32)&Z80CORESTART);

	// allocate 7K for jump tables
	jtBase=(uint32*)MemAlloc(HEAP_EXTERNAL,7*1024);
	ASSERT(jtBase);
	memcpy(jtBase,&JT_MAIN,7*1024);
	z80Ptr->mainTable=jtBase;

	// fill in missing ptrs
	jtBase[0xCB]=(uint32)&DoCB;
	jtBase[0xED]=(uint32)&DoED;
	jtBase[0xDD]=(uint32)&DoDD;
	jtBase[0xFD]=(uint32)&DoFD;
	jtBase[3*256 + 0xCB]=(uint32)&DoDDCB;
	jtBase[4*256 + 0xCB]=(uint32)&DoFDCB;

	// patch up tables
	for(i=0;i<7*256;i++)
	{
		if(!jtBase[i])	jtBase[i]=(uint32)(&DoBAD);
		jtBase[i] += ((uint32)z80Base) - ((uint32)&Z80CORESTART);
	}

	// copy first 1K (non-prefixed instructions) into fast RAM also
	jtFast=(uint32*)MemAlloc(HEAP_INTERNAL,1024);
	ASSERT(jtFast);
	mymemcpy(jtFast,jtBase,1024);
	z80Ptr->fastMainTable=jtFast;

	// set flag tables
	z80Ptr->tables=&rotateFlagTable[0];

	// allocate z80 RAM
	z80Mem=(byte*)MemAlloc(HEAP_EXTERNAL,65536+16);
	ASSERT(z80Mem);
	for(i=0;i<65536+16;i++)	z80Mem[i]=0;
	z80Ptr->state.mem=z80Mem;

	// copy 16K spectrum rom
	memcpy(z80Mem,specRom,16384);

	// repeat 16 bytes at end of memory!
	for(i=0;i<16;i++)	z80Mem[i+65536]=z80Mem[i];

	// allocate port vals
	ports=(byte *)MemAlloc(HEAP_EXTERNAL,512);
	ASSERT(ports);
	z80Ptr->feInputPtr=&ports[0];
	z80Ptr->xxInputPtr=&ports[256];

	// shadow init
#ifdef SHADOW
	shadowMem=(byte *)MemAlloc(HEAP_EXTERNAL,65536);
	ASSERT(shadowMem);
	ResetZ80(&shadowZ80);
	shadowZ80.SP.W=0;
	EmuCopyState();
#endif

	// force port init
	lastKey=-1;
	EmuUpdatePorts();

	// allocate backup state/mem
	bkpMem=(byte*)MemAlloc(HEAP_EXTERNAL,49152);
	ASSERT(bkpMem);
	if(bkpMem)	EmuBackup();

}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

void EmuCopyScreen(void)
{
	int32		f;

	// FLASH?
	z80Ptr->flashCt=768;
	f=frameCt/25;
	if(f&1)
	{
		z80Ptr->flashPtr=&z80Mem[0x5800];
		Call32(flashEntry);
	}

	DoCopyup();

	// FLASH?
	if(f&1)
	{
		z80Ptr->flashPtr=&z80Mem[0x5800];
		Call32(flashEntry);
	}

	frameCt++;
}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

void EmuCopyThirdScreen(int32 which)
{
	int32		f;

	// FLASH?
	z80Ptr->flashCt=256;
	f=frameCt/25;
	if(f&1)
	{
		z80Ptr->flashPtr=&z80Mem[0x5800 + (which<<8)];
		Call32(flashEntry);
	}

	DoThirdCopyup(which);

	// FLASH?
	if(f&1)
	{
		z80Ptr->flashPtr=&z80Mem[0x5800 + (which<<8)];
		Call32(flashEntry);
	}

	if(!which)	frameCt++;
}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

void DumpState(void)
{
	DrawString("AF",0,  8,0,255);		DrawHex(*((uint16*)&z80Ptr->state.F),24, 8,0,253);
	DrawString("BC",0, 18,0,255);		DrawHex(*((uint16*)&z80Ptr->state.C),24,18,0,253);
	DrawString("DE",0, 28,0,255);		DrawHex(*((uint16*)&z80Ptr->state.E),24,28,0,253);
	DrawString("HL",0, 38,0,255);		DrawHex(*((uint16*)&z80Ptr->state.L),24,38,0,253);
	DrawString("IX",0, 48,0,255);		DrawHex(z80Ptr->state.IX,24,48,0,253);
	DrawString("IY",0, 58,0,255);		DrawHex(z80Ptr->state.IY,24,58,0,253);
	DrawString("SP",0, 68,0,255);		DrawHex(z80Ptr->state.SP,24,68,0,253);
	DrawString("PC",0, 78,0,255);		DrawHex(z80Ptr->state.PC,24,78,0,253);
	DrawString("AF-",0, 88,0,255);		DrawHex(*((uint16*)&z80Ptr->state.F_),24, 88,0,253);
	DrawString("BC-",0, 98,0,255);		DrawHex(*((uint16*)&z80Ptr->state.C_),24, 98,0,253);
	DrawString("DE-",0,108,0,255);		DrawHex(*((uint16*)&z80Ptr->state.E_),24,108,0,253);
	DrawString("HL-",0,118,0,255);		DrawHex(*((uint16*)&z80Ptr->state.L_),24,118,0,253);
	DrawString("RI",0,128,0,255);		DrawHex(((int32)z80Ptr->state.R<<8) | z80Ptr->state.I,24,128,0,253);

#ifdef SHADOW
	DrawString("AF",120,  8,0,255);		DrawHex(shadowZ80.AF.W,144,  8,0,253);
	DrawString("BC",120, 18,0,255);		DrawHex(shadowZ80.BC.W,144, 18,0,253);
	DrawString("DE",120, 28,0,255);		DrawHex(shadowZ80.DE.W,144, 28,0,253);
	DrawString("HL",120, 38,0,255);		DrawHex(shadowZ80.HL.W,144, 38,0,253);
	DrawString("IX",120, 48,0,255);		DrawHex(shadowZ80.IX.W,144, 48,0,253);
	DrawString("IY",120, 58,0,255);		DrawHex(shadowZ80.IY.W,144, 58,0,253);
	DrawString("SP",120, 68,0,255);		DrawHex(shadowZ80.SP.W,144, 68,0,253);
	DrawString("PC",120, 78,0,255);		DrawHex(shadowZ80.PC.W,144, 78,0,253);
	DrawString("AF-",120, 88,0,255);	DrawHex(shadowZ80.AF1.W,144, 88,0,253);
	DrawString("BC-",120, 98,0,255);	DrawHex(shadowZ80.BC1.W,144, 98,0,253);
	DrawString("DE-",120,108,0,255);	DrawHex(shadowZ80.DE1.W,144,108,0,253);
	DrawString("HL-",120,118,0,255);	DrawHex(shadowZ80.HL1.W,144,118,0,253);
	DrawString("I",120,128,0,255);		DrawHex(shadowZ80.I,144,128,0,253);
#endif

	DrawHex(lastPC,0,140,0,253);
	DrawHex(lastOp,60,140,0,253);
	DrawHex(lastOp2,92,140,0,253);
	DrawHex(lastFlags,120,140,0,253);

	DrawString("IM",180,8,0,255);
	DrawHex(z80Ptr->state.IM,200,8,0,255);
	DrawString("IF",180,18,0,255);
	DrawHex(z80Ptr->state.IFF,200,18,0,255);

	if(memFailure)
	{
		DrawHex(memFailAddr,180,140,0,255);
	}
}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

void EmuTrace(void)
{
	int32		i;
	byte		dummy;

	CopyupSetMode(CM_LINEAR);

	z80Ptr->state.rInc=0;
	i=0;
	trace=false;
	while(1)
	{
		lastPC=z80Ptr->state.PC;
		lastOp=(uint32)z80Mem[lastPC]<<8;
		lastOp|=z80Mem[lastPC+1];
		lastOp2=(uint32)z80Mem[lastPC+2]<<8;
		lastOp2|=z80Mem[lastPC+3];
		lastFlags=z80Ptr->state.F;

		z80Ptr->state.clocks=0;
		z80Ptr->state.samplesPerFrame=1;
		z80Ptr->state.cyclesPerSample=0;
		z80Ptr->state.audioPtr=&dummy;

		z80Ptr->state.R=0x0a;
		Call32(z80Entry);
		ASSERT(!z80Ptr->state.badInst);

#ifdef SHADOW
		shadowZ80.ICount=0x7fffffff;
		ExecZ80(&shadowZ80);
#endif

		if((i++)==AUDIO_CYCLESPERFRAME || singleStep)
		{
			EmuCopyScreen();
			i=0;
			*((uint32*)VRAM)=i;
			DumpState();
			EmuUpdatePorts();
			if(singleStep)	WaitKey();

			EmuDoInt();
			lastOp=0xDEADBEEF;
		}

#ifdef SHADOW
		if(CheckState())
		{
			DumpState();
			*((uint32*)VRAM)=0xFFFF;
			while(1);
		}


		// overwrite flags!!!
		shadowZ80.AF.B.l=z80Ptr->state.F;
#endif

		if(trace)
		{
			DrawHex(z80Ptr->state.PC,60,5,0,11);
		}
	}
}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

void EmuRun(void)
{
	int32		i,li,th;
	int32		ab;

	// init Z80 values
	z80Ptr->state.clocks=0;
	z80Ptr->state.rInc=imageRinc;
	ab=0;
	
	i=0; li=0;
	while(1)
	{
		// update keyboard stuff
		EmuUpdatePorts();

		// run emulator
		z80Ptr->state.audioPtr=audioBuffers[ab];
		z80Ptr->state.cyclesPerSample=AUDIO_CYCLESPERSAMPLE;
		z80Ptr->state.badInst=false;

		// do 3 sections of screen
		for(th=0;th<3;th++)
		{
#if 1
	#ifdef USE_AUDIO
			z80Ptr->state.clocks+=AUDIO_CYCLESPERSAMPLE;
			if(imageScreenMode!=3)
			{
				z80Ptr->state.samplesPerFrame=AUDIO_SAMPLESPERFRAME / 3;

				// doh - have to have 320 samples, but this doesn't divide by 3...
				// so we add an extra sample to the first two 1/3rds of the screen
				if(th!=2)	z80Ptr->state.samplesPerFrame++;
			}
			else
			{
				// we're doing a full frame
				z80Ptr->state.samplesPerFrame=AUDIO_SAMPLESPERFRAME;
			}
	#else
			byte		dummy;
			if(imageScreenMode!=3)		z80Ptr->state.clocks+=AUDIO_CYCLESPERFRAME / 3;
			else				z80Ptr->state.clocks+=AUDIO_CYCLESPERFRAME;	// 70000ish
			z80Ptr->state.samplesPerFrame=1;
			z80Ptr->state.audioPtr=&dummy;
	#endif
			Call32(z80Entry);
			ASSERT(!z80Ptr->state.badInst);
#else
			for(i=0;i<AUDIO_CYCLESPERFRAME/18;i++) ExecZ80(&shadowZ80);
			memcpy(&z80Mem[0x4000],&shadowMem[0x4000],6912);
#endif

			// fullscreen mode?
			if(imageScreenMode==3)
			{
				if(frameSkipCt>=frameSkip)
				{
					EmuCopyScreen();
				}
				EmuDoInt();
				break;
			}

			// frameskip?
			if(frameSkipCt >= frameSkip)
			{
				// copy up screen
				EmuCopyThirdScreen(th);
			}

			if(th==imageScreenMode)	EmuDoInt();
		}

		// increment frame skip ctr
		if(frameSkipCt>=frameSkip)	frameSkipCt=0;
		else				frameSkipCt++;
	
		i++;

#ifdef USE_AUDIO
		// wait for audio complete, play next buffer
		IntSubmitBuffer(ab);
		ab^=1;
#endif
		// break out if START pressed...
		if(lastKey & 8)
		{
			EmuCopyState();
			return;
		}
	}
}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

int32 EmuReadByte(byte **ptr)
{
	return(*((*ptr)++));
}
int32 EmuReadWord(byte **ptr)
{
	int32		w;
	w =EmuReadByte(ptr);
	w|=EmuReadByte(ptr)<<8;
	return(w);
}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

int32 EmuLoadZ80(Header *img,bool fetchHeaderInfo)
{
	int32		i;
	byte		*orig;
	byte		*ptr;
	int32		imgLen;
	int32		ver;
	int32		flg;

	ver=-1;

	// do we want to overwrite the current settings?
	if(fetchHeaderInfo)
	{
		// take copy for keydefs
		memcpy(keys,img->keys,sizeof(keys));

		// copy flags
		imageFlags=img->flags;
		imageScreenMode=(int32)((imageFlags>>HFLAGS_SCR_SHIFT) & HFLAGS_SCR_MASK);
		imageRinc=(int32)((imageFlags>>HFLAGS_RINC_SHIFT) & HFLAGS_RINC_MASK);
	}

	ptr=(byte*)(&img[1]);
	imgLen=img->len;

	orig=ptr;

	mymemset(&z80Ptr->state,0,sizeof(z80Ptr->state));
	z80Ptr->state.mem=z80Mem;

	z80Ptr->state.A=EmuReadByte(&ptr);
	z80Ptr->state.F=EmuReadByte(&ptr);
	z80Ptr->state.C=EmuReadByte(&ptr);		z80Ptr->state.B=EmuReadByte(&ptr);
	z80Ptr->state.L=EmuReadByte(&ptr);		z80Ptr->state.H=EmuReadByte(&ptr);
	
	z80Ptr->state.PC=EmuReadWord(&ptr);
	if(z80Ptr->state.PC)	ver=0;

	z80Ptr->state.SP=EmuReadWord(&ptr);
	z80Ptr->state.I=EmuReadByte(&ptr);
	z80Ptr->state.R=EmuReadByte(&ptr);
	z80Ptr->state.border=EmuReadByte(&ptr);
	flg=z80Ptr->state.border;

	z80Ptr->state.E=EmuReadByte(&ptr);		z80Ptr->state.D=EmuReadByte(&ptr);

	z80Ptr->state.C_=EmuReadByte(&ptr);		z80Ptr->state.B_=EmuReadByte(&ptr);
	z80Ptr->state.E_=EmuReadByte(&ptr);		z80Ptr->state.D_=EmuReadByte(&ptr);
	z80Ptr->state.L_=EmuReadByte(&ptr);		z80Ptr->state.H_=EmuReadByte(&ptr);
	z80Ptr->state.A_=EmuReadByte(&ptr);		z80Ptr->state.F_=EmuReadByte(&ptr);

	z80Ptr->state.IY=EmuReadWord(&ptr);
	z80Ptr->state.IX=EmuReadWord(&ptr);
	z80Ptr->state.IFF=0;
	if(EmuReadByte(&ptr))	z80Ptr->state.IFF |= IFF_1;
	if(EmuReadByte(&ptr))	z80Ptr->state.IFF |= IFF_2;
	z80Ptr->state.IM=EmuReadByte(&ptr) & 3;

	if(ver!=0)
	{
		// get extra header length
		i=EmuReadWord(&ptr);
		if(i==23)	ver=2;
		else if(i==54)	ver=3;
		else
		{
			return(-3);
		}

		// get PC
		z80Ptr->state.PC=EmuReadWord(&ptr);

		// check mode
		i=EmuReadByte(&ptr);	if(i!=0 && i!=1) { return(-4); }

		EmuReadByte(&ptr);
		i=EmuReadByte(&ptr);	if(i==0xFF) { return(-5); }

		EmuReadByte(&ptr);
		EmuReadByte(&ptr);
		for(i=0;i<16;i++) EmuReadByte(&ptr);

		// version 3 - skip extra
		if(ver==3)	for(i=0;i<54-23;i++) EmuReadByte(&ptr);
	}

	while(1)
	{
		int32		pageLen,pageNum,pageOff;
		int32		wPtr,state,rept,val,len;

		if((ptr-orig) >= imgLen)	break;
		if(ver==0 && (flg&32))
		{
			pageOff=0x4000;
			pageLen=imgLen - (ptr-orig);
		}
		else
		{
			pageLen=EmuReadByte(&ptr);		pageLen+=EmuReadByte(&ptr)<<8;
			if((ptr-orig) >= imgLen)	break;
			pageNum=EmuReadByte(&ptr);
			if(pageNum==0)		pageOff=0x0000;
			else if(pageNum==4)	pageOff=0x8000;
			else if(pageNum==5)	pageOff=0xC000;
			else if(pageNum==8)	pageOff=0x4000;
			else
			{
				return(-6);
			}
		}

		wPtr=pageOff;
		len=0;
		state=0;
		rept=0;
		while(len<pageLen)
		{
			i=EmuReadByte(&ptr);	len++;

			if(state==0)
			{
				if(i!=0xED)	z80Ptr->state.mem[wPtr++]=(byte)i;
				else		state=1;
			}
			else if(state==1)
			{
				if(i!=0xED)	{ z80Ptr->state.mem[wPtr++]=(byte)0xED; z80Ptr->state.mem[wPtr++]=(byte)i; state=0; }
				else		state=2;
			}
			else if(state==2)
			{
				rept=i;
				state=3;
			}
			else if(state==3)
			{
				val=i;
				while(rept--)	z80Ptr->state.mem[wPtr++]=(byte)val;
				state=0;
			}
		}
	}

	// repeat 16 bytes at end of memory!
	for(i=0;i<16;i++)	z80Mem[i+65536]=z80Mem[i];

	// copy state
	EmuCopyState();

	return(0);
}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

void EmuUpdatePorts(void)
{
	int32		i,j;
	int32		key;

	// read pad
	key=*((volatile uint16*)REG_P1) ^ 0xFFFF;
	
	// return if same
	if(key==lastKey && !injectedKey)	return;

	// store new value
	lastKey=key;

	// clear ports
	for(i=0;i<512;i++) ports[i]=0xFF;
	ports[0x11f]=0x00;

	// clear keys
	for(i=0;i<8;i++)	keyboard[i]=0xFF;

	// generate keypresses
	if(!(key&256) || copyupMode!=CM_TILED)			// disable R shoulder in scrolling mode
	{
		for(i=0,j=32768;i<16;i++,j>>=1)
		{
			if(key & j)
			{
				if(keys[i].type==KEY_JOY)
				{
					ports[ keys[i].port | 0x100 ] |= keys[i].val;
				}
				else if(keys[i].type==KEY_KBD)
				{
					keyboard[ keys[i].port ] &= ~keys[i].val;
				}
			}
		}
	}

	// check injected keys
	if(injectedKey)
	{
		j=0;
		for(i=0;i<keyCt;i++)
		{
			if(keyPress[i])
			{
				if(keyDefs[i].k.type==KEY_KBD)
				{
					keyboard[ keyDefs[i].k.port ] &= ~keyDefs[i].k.val;
				}
				else
				{
					ports[ keyDefs[i].k.port | 0x100 ] |= keyDefs[i].k.val;
				}
				keyPress[i]--;
				j=1;
			}
		}
		if(!j)	injectedKey=0;
	}

	// generate keyboard bytes
	z80Ptr->keyPortPtr=ports;
	z80Ptr->keyPtr=keyboard;
	Call32(bkmEntry);

}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

void EmuDoInt(void)
{
	uint16		sp,pc;
	int32		i;

	// if interrupts are disabled, we just skip this frame
	// strictly, we should queue this interrupt, but then again
	// if the Z80's still in an int after the previous frame,
	// forget it!
	if(!(z80Ptr->state.IFF & IFF_1))	return;

	// unhalt (halt flag gets reset further down)
	if(z80Ptr->state.IFF & HALT)
	{
		z80Ptr->state.PC++;
	}

	// push program counter
	sp=z80Ptr->state.SP;
	pc=z80Ptr->state.PC;
	sp--;
	if(sp>=0x4000)	z80Mem[sp]=(byte)(pc>>8);
	sp--;
	if(sp>=0x4000)	z80Mem[sp]=(byte)(pc&0xFF);
	z80Ptr->state.SP=sp;

	// reset IFFs (and halt flag)
	z80Ptr->state.IFF=0;

	// process interrupt depending on mode
	if(z80Ptr->state.IM==0)
	{
		z80Ptr->state.PC=0x38;
	}
	else if(z80Ptr->state.IM==1)
	{
		z80Ptr->state.PC=0x38;
	}
	else if(z80Ptr->state.IM==2)
	{
		i =((uint16)z80Ptr->state.I)<<8;
		i|=0xFF;
		pc =z80Mem[i];
		pc|=((int32)z80Mem[(i+1)&0xFFFF])<<8;

		z80Ptr->state.PC=pc;
		intCt++;
	}

#ifdef SHADOW
	IntZ80(&shadowZ80,0x38);
#endif
}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

void EmuCopyState(void)
{
#ifdef SHADOW
	shadowZ80.AF.B.h=z80Ptr->state.A;
	shadowZ80.AF.B.l=z80Ptr->state.F;
	shadowZ80.BC.B.h=z80Ptr->state.B;
	shadowZ80.BC.B.l=z80Ptr->state.C;
	shadowZ80.DE.B.h=z80Ptr->state.D;
	shadowZ80.DE.B.l=z80Ptr->state.E;
	shadowZ80.HL.B.h=z80Ptr->state.H;
	shadowZ80.HL.B.l=z80Ptr->state.L;
	shadowZ80.PC.W=z80Ptr->state.PC;
	shadowZ80.IX.W=z80Ptr->state.IX;
	shadowZ80.IY.W=z80Ptr->state.IY;
	shadowZ80.SP.W=z80Ptr->state.SP;
	shadowZ80.AF1.B.h=z80Ptr->state.A_;
	shadowZ80.AF1.B.l=z80Ptr->state.F_;
	shadowZ80.BC1.B.h=z80Ptr->state.B_;
	shadowZ80.BC1.B.l=z80Ptr->state.C_;
	shadowZ80.DE1.B.h=z80Ptr->state.D_;
	shadowZ80.DE1.B.l=z80Ptr->state.E_;
	shadowZ80.HL1.B.h=z80Ptr->state.H_;
	shadowZ80.HL1.B.l=z80Ptr->state.L_;
	shadowZ80.I=z80Ptr->state.I;

	shadowZ80.IFF =(z80Ptr->state.IFF & IFF_1)?0x01:0;
	shadowZ80.IFF|=(z80Ptr->state.IFF & IFF_2)?0x40:0;
	shadowZ80.IFF|=(z80Ptr->state.IFF & HALT)?0x80:0;
	shadowZ80.IFF|=(z80Ptr->state.IM == 1)?0x02:0;
	shadowZ80.IFF|=(z80Ptr->state.IM == 2)?0x04:0;

	memcpy(shadowMem,z80Mem,65536);
#endif
}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

char	sig2[]="GS!!";
char	xxxx[]="xx";
char	sig1[]="STRA";

int32 EmuFindImage(int32 start)
{
	int32		i; 
	int32		isig1,isig2;

	isig1=*((int32*)&sig1[0]);
	isig2=*((int32*)&sig2[0]);

	start&=~3;
	for(i=start;i<0x08000000 | (8*1024*1024);i+=4)
	{
		if( (*((int32*)i)) == isig1 )
		{
			if( (*((int32*)(i+4))) == isig2 )
			{
				return(i);
			}
		}
	}

	return(0);
}

int32 EmuGotSig(int32 ad)
{
	int32		isig1,isig2;

	isig1=*((int32*)&sig1[0]);
	isig2=*((int32*)&sig2[0]);

	if( (*((int32*)ad)) == isig1 )
	{
		if( (*((int32*)(ad+4))) == isig2 )
		{
			return(1);
		}
	}

	return(0);
}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

void EmuBackup(void)
{
	if(!bkpMem)	return;
	mymemcpy(bkpMem,&z80Mem[16384],49152);
	mymemcpy(&bkpState,&z80Ptr->state,sizeof(bkpState));
	mymemcpy(bkpKeys,keys,sizeof(bkpKeys));
}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

void EmuRestore(void)
{
	if(!bkpMem)	return;
	mymemcpy(z80Mem,specRom,16384);
	mymemcpy(&z80Mem[16384],bkpMem,49152);
	mymemcpy(&z80Ptr->state,&bkpState,sizeof(bkpState));
	mymemcpy(keys,bkpKeys,sizeof(bkpKeys));
	z80Ptr->state.mem=z80Mem;
}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

#define DIFF(a,b) ((a)!=(b))
int32 CheckState(void)
{
#ifdef SHADOW
	if(memFailure)	return(-2);
	if(DIFF(shadowZ80.AF.B.h,z80Ptr->state.A))	return(-1);
	if(DIFF(shadowZ80.AF.B.l&FLG_CARE,z80Ptr->state.F&FLG_CARE))	return(-1);
	if(DIFF(shadowZ80.BC.B.h,z80Ptr->state.B))	return(-1);
	if(DIFF(shadowZ80.BC.B.l,z80Ptr->state.C))	return(-1);
	if(DIFF(shadowZ80.DE.B.h,z80Ptr->state.D))	return(-1);
	if(DIFF(shadowZ80.DE.B.l,z80Ptr->state.E))	return(-1);
	if(DIFF(shadowZ80.HL.B.h,z80Ptr->state.H))	return(-1);
	if(DIFF(shadowZ80.HL.B.l,z80Ptr->state.L))	return(-1);
	if(DIFF(shadowZ80.PC.W,z80Ptr->state.PC))	return(-1);
	if(DIFF(shadowZ80.IX.W,z80Ptr->state.IX))	return(-1);
	if(DIFF(shadowZ80.IY.W,z80Ptr->state.IY))	return(-1);
	if(DIFF(shadowZ80.SP.W,z80Ptr->state.SP))	return(-1);
	if(DIFF(shadowZ80.AF1.B.h,z80Ptr->state.A_))	return(-1);
	if(DIFF(shadowZ80.AF1.B.l,z80Ptr->state.F_))	return(-1);
	if(DIFF(shadowZ80.BC1.B.h,z80Ptr->state.B_))	return(-1);
	if(DIFF(shadowZ80.BC1.B.l,z80Ptr->state.C_))	return(-1);
	if(DIFF(shadowZ80.DE1.B.h,z80Ptr->state.D_))	return(-1);
	if(DIFF(shadowZ80.DE1.B.l,z80Ptr->state.E_))	return(-1);
	if(DIFF(shadowZ80.HL1.B.h,z80Ptr->state.H_))	return(-1);
	if(DIFF(shadowZ80.HL1.B.l,z80Ptr->state.L_))	return(-1);
	if(DIFF(shadowZ80.I,z80Ptr->state.I))		return(-1);

	shadowZ80.IFF =(z80Ptr->state.IFF & IFF_1)?0x01:0;
	shadowZ80.IFF|=(z80Ptr->state.IFF & IFF_2)?0x40:0;
	shadowZ80.IFF|=(z80Ptr->state.IFF & HALT)?0x80:0;
	shadowZ80.IFF|=(z80Ptr->state.IM == 1)?0x02:0;
	shadowZ80.IFF|=(z80Ptr->state.IM == 2)?0x04:0;
	if(DIFF((shadowZ80.IFF&0x01)?1:0,(z80Ptr->state.IFF & IFF_1)?1:0))	return(-3);
	if(DIFF((shadowZ80.IFF&0x40)?1:0,(z80Ptr->state.IFF & IFF_2)?1:0))	return(-3);
	if(DIFF((shadowZ80.IFF&0x80)?1:0,(z80Ptr->state.IFF & HALT)?1:0))	return(-3);
	if(DIFF((shadowZ80.IFF>>1)&3,z80Ptr->state.IM))				return(-3);
#endif
	return(0);
}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/
#ifdef SHADOW
void WrZ80(register word Addr,register byte Value)
{
	if(Addr>=0x4000)
	{
		shadowMem[Addr]=Value;
	}
	if(shadowMem[Addr] != z80Mem[Addr])
	{
		memFailure=true;
		memFailAddr=Addr;
	}
}
byte RdZ80(register word Addr)
{
	return(shadowMem[Addr]);
}

/** InZ80()/OutZ80() *****************************************/
/** Z80 emulation calls these functions to read/write from  **/
/** I/O ports. There can be 65536 I/O ports, but only first **/
/** 256 are usually used                                    **/
/************************************ TO BE WRITTEN BY USER **/
void OutZ80(register word Port,register byte Value)
{
}
byte InZ80(register word Port)
{
	if((Port&0xFF)==0xFE)
	{
		return(ports[(Port>>8)]);
	}
	else
	{
		return(ports[0x100+(Port&0xFF)]);
	}
}

/** PatchZ80() ***********************************************/
/** Z80 emulation calls this function when it encounters a  **/
/** special patch command (ED FE) provided for user needs.  **/
/** For example, it can be called to emulate BIOS calls,    **/
/** such as disk and tape access. Replace it with an empty  **/
/** macro for no patching.                                  **/
/************************************ TO BE WRITTEN BY USER **/
void PatchZ80(register Z80 *R)
{
}

word LoopZ80(register Z80 *R)
{
	return(INT_NONE);
}

#endif
