/***********************************************************************
 *
 * sram.cpp
 *
 ***********************************************************************
 *
 * sram load/save
 *
 ***********************************************************************/

/*
 * includes
 */

#include "std.h"
#include "sram.h"
#include "emu.h"
#include "main.h"
#include "shell.h"
#include "gba.h"
#include "mem.h"
#include "font.h"

/*
 * defines
 */

#define SRAM_MAX	1024*1024

/*
 * typedefs
 */

/*
 * structs
 */

/*
 * globals
 */

int32		sramSize;
int32		sramSnapshotsPresent;
uint16		wsBkp;
volatile byte*	sramPtr=(byte *)0x0e000000;

uint32		sramChk;
uint32		lastChk;
int32		sramCt;
int32		sramOff;
bool		sramVerifyFailed;

char		*sramSig="FOON-BY-STRAGS!!";

/*
 * function prototypes
 */

void		SRAMStart(void);
void		SRAMStop(void);
void		SRAMWriteByte(byte c);
int32		SRAMVerifyByte(byte c);
byte		SRAMReadByte(void);

/*
 * code
 */

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

void ChkAdd(int32 off,int32 c)
{
	c&=0xFF;
	off&=7;
	c<<=off;
	sramChk ^= c;
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

void SRAMStart(void)
{
	wsBkp=*((uint16*)0x02000204);
	*((uint16*)0x02000204) = wsBkp|3;	// 8-wait states for SRAM
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

void SRAMStop(void)
{
	*((uint16*)0x02000204) = wsBkp;
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

void SRAMScan(void)
{
	int32		i;
	byte		b1,b2,b3;
	volatile byte	*p;

	// prepare waitstates
	SRAMStart();

	// get ptr
	p=sramPtr;
	b2=0xFF;

	// test wrapping and writeability
	b1=p[0];
	for(i=0;i<SRAM_MAX;i+=256)
	{
		b3=p[i];
		p[i]=(byte)(b3^b2);
		if(p[i]!=(byte)(b3^b2))		break;
		if(b1!=p[0])
		{
			if(i)	break;
		}
		p[i]=b3;
	}
	if(i!=SRAM_MAX)	p[i]=b3;

	sramSize=i;

	// restore waitstates
	SRAMStop();
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

int32 SRAMSave(void)
{
	int32		i,j;

	// init waitstates
	SRAMStart();

	// init writer
	sramCt=0;
	sramOff=64;
	sramChk=0x1234;

	// write signature
	for(i=0;i<16;i++)	SRAMWriteByte(sramSig[i]);

	// version
	SRAMWriteByte(SRAM_VERSION&0xFF);
	SRAMWriteByte(SRAM_VERSION>>8);

	// keys
	for(i=0;i<64;i++)		SRAMWriteByte(((byte*)(&keys[0]))[i]);

	// state
	for(i=0;i<(int32)sizeof(z80Ptr->state);i++)	SRAMWriteByte(((byte*)(&z80Ptr->state))[i]);

	// memory
	for(i=16384;i<65536;i++)
	{
		if(!(i&0xFF))
		{
			for(j=0;j<8;j++)
			{
				((uint16*)VRAM)[j*120+(((i-16384)>>9))+12+0x0000]=0x1818;
				((uint16*)VRAM)[j*120+(((i-16384)>>9))+12+0x5000]=0x1818;
			}
		}
		SRAMWriteByte(z80Mem[i]);
	}

	// write checksum
	sramChk &= 0xffff;
	j=sramChk;
	lastChk=sramChk;
	SRAMWriteByte(j&0xFF);
	SRAMWriteByte(j>>8);

	// done with waitstates
	SRAMStop();

	// verify!
	if(SRAMVerify())	return(-1);

	return(0);
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

int32 SRAMVerify(void)
{
	int32		i,j;
	
	// init waitstates
	SRAMStart();

	// init verifier
	sramCt=0;
	sramOff=64;
	sramChk=0x1234;
	sramVerifyFailed=false;

	// write signature
	for(i=0;i<16;i++)	SRAMVerifyByte(sramSig[i]);
	if(sramVerifyFailed)	goto	err;

	// version
	SRAMVerifyByte(SRAM_VERSION&0xFF);
	SRAMVerifyByte(SRAM_VERSION>>8);
	if(sramVerifyFailed)	goto	err;

	// keys
	for(i=0;i<64;i++)		SRAMVerifyByte(((byte*)(&keys[0]))[i]);
	if(sramVerifyFailed)	goto	err;

	// state
	for(i=0;i<(int32)sizeof(z80Ptr->state);i++)	SRAMVerifyByte(((byte*)(&z80Ptr->state))[i]);
	if(sramVerifyFailed)	goto	err;

	// memory
	for(i=16384;i<65536;i++)
	{
		if(!(i&0xFF))
		{
			for(j=0;j<8;j++)
			{
				((uint16*)VRAM)[j*120+(((i-16384)>>9))+12+0x0000]=0x0000;
				((uint16*)VRAM)[j*120+(((i-16384)>>9))+12+0x5000]=0x0000;
			}
		}
		SRAMVerifyByte(z80Mem[i]);
	}
	if(sramVerifyFailed)	goto	err;

	// write checksum
	sramChk &= 0xffff;
	j=sramChk;
	SRAMVerifyByte(j&0xFF);
	SRAMVerifyByte(j>>8);

err:
	// done with waitstates
	SRAMStop();

	// check flag
	if(sramVerifyFailed)	return(-1);

	return(0);
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

void SRAMWriteByte(byte c)
{
	sramPtr[sramCt+sramOff]=c;
	ChkAdd(sramCt,c);
	sramCt++;
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

int32 SRAMVerifyByte(byte c)
{
	if(sramPtr[sramCt+sramOff]!=c)	sramVerifyFailed=true;
	ChkAdd(sramCt,c);
	sramCt++;

	if(sramVerifyFailed)	return(-1);
	return(0);
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

byte SRAMReadByte(void)
{
	byte		c;
	c=sramPtr[sramCt+sramOff];
	ChkAdd(sramCt,c);
	sramCt++;
	return(c);
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

int32 SRAMLoad(void)
{
	int32		i,j;

	// init waitstates
	SRAMStart();

	// init writer
	sramCt=0;
	sramOff=64;
	sramChk=0x1234;
	sramVerifyFailed=false;

	// verify signature
	for(i=0;i<16;i++)	SRAMVerifyByte(sramSig[i]);
	if(sramVerifyFailed)	goto err;

	// version
	SRAMVerifyByte(SRAM_VERSION&0xFF);
	SRAMVerifyByte(SRAM_VERSION>>8);
	if(sramVerifyFailed)	goto err;

	// keys
	for(i=0;i<64;i++)				((byte*)(&keys[0]))[i]=SRAMReadByte();

	// state
	for(i=0;i<(int32)sizeof(z80Ptr->state);i++)	((byte*)(&z80Ptr->state))[i]=SRAMReadByte();

	// memory
	for(i=16384;i<65536;i++)
	{
		if(!(i&0xFF))
		{
			for(j=0;j<8;j++)
			{
				((uint16*)VRAM)[j*120+(((i-16384)>>9))+12+0x0000]=0x1818;
				((uint16*)VRAM)[j*120+(((i-16384)>>9))+12+0x5000]=0x1818;
			}
		}
		z80Mem[i]=SRAMReadByte();
	}

	// check checksum
	sramChk &= 0xffff;
	j=sramChk;

	SRAMVerifyByte(j&0xFF);
	SRAMVerifyByte(j>>8);

	// done with waitstates
err:
	SRAMStop();

	// restore internal ptrs
	z80Ptr->state.mem=z80Mem;

	// verify!
	if(sramVerifyFailed)	return(-1);

	return(0);
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

int32 SRAMWriteOptions(void)
{
	int32		i,j;

	return(-1);

	// init
	SRAMStart();
	sramCt=0;
	sramOff=0;
	sramChk=0x1234;

	// write signature
	for(i=0;i<16;i++)	SRAMWriteByte(sramSig[i]);

	// version
	SRAMWriteByte(SRAM_VERSION&0xFF);
	SRAMWriteByte(SRAM_VERSION>>8);

	// data
	SRAMWriteByte(videoMode);
	SRAMWriteByte(brightness);
	SRAMWriteByte(colourMode);
	SRAMWriteByte(frameSkip);

	// write checksum
	j=sramChk;
	SRAMWriteByte(j&0xFF);
	SRAMWriteByte(j>>8);

	// done
	SRAMStop();
	return(0);
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

int32 SRAMReadOptions(void)
{
	int32		i,j;

	return(-1);

	// init
	SRAMStart();
	sramCt=0;
	sramOff=0;
	sramChk=0x1234;
	sramVerifyFailed=false;

	// write signature
	for(i=0;i<16;i++)	SRAMVerifyByte(sramSig[i]);
	if(sramVerifyFailed)	goto err;

	// version
	SRAMVerifyByte(SRAM_VERSION&0xFF);
	SRAMVerifyByte(SRAM_VERSION>>8);
	if(sramVerifyFailed)	goto err;

	// data
	videoMode=SRAMReadByte();
	brightness=SRAMReadByte();
	colourMode=SRAMReadByte();
	frameSkip=SRAMReadByte();

	// checksum
	j=sramChk;
	SRAMVerifyByte(j&0xFF);
	SRAMVerifyByte(j>>8);

err:
	SRAMStop();

	if(sramVerifyFailed)	return(-1);
	return(0);
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

